#%% init project env
import esypro
import scipy.sparse
sfs = esypro.ScriptResultManager('zqf',locals(), version=0)

from . import basic
import torch
import scipy


class CSRMatrixDataset(basic.IterableData):
    """
    Iterate over a CSR matrix.
    The dataset will return a dense vector for each index.
    """
    def __init__(self, csr_matrix:scipy.sparse._csr.csr_matrix, sample_selection=None, **kwargs):
        self.indicies = torch.tensor(csr_matrix.indices)
        self.indptrs = torch.tensor(csr_matrix.indptr)
        self.data = torch.tensor(csr_matrix.data)
        self.shape = csr_matrix.shape
        self.line_index = torch.arange(len(self.indptrs)-1)
        
        super().__init__(self.line_index, sample_selection=sample_selection, **kwargs)
        
    def __getitem__(self, select_index):
        index = super().__getitem__(select_index)
        x_values = self.data[self.indptrs[index]:self.indptrs[index+1]]
        x_indexs = self.indicies[self.indptrs[index]:self.indptrs[index+1]]
        x_pointers = self.indptrs[index:index+2] - self.indptrs[index]
        csr_matrix = scipy.sparse.csr_matrix((x_values, x_indexs, x_pointers), shape=(1, self.shape[1]))
        dense_vector = torch.tensor(csr_matrix.todense()).squeeze(0)
        return dense_vector
        
    
    

#%% main
if __name__ == '__main__':
    print(f'start {__file__}')

    data = scipy.sparse.csr_matrix([[1, 0, 0], [0, 2, 0], [3, 0, 4]])
    dataset = CSRMatrixDataset(data)
    
    #%%
    dataset[0]
    
    #%% end script
    print(f'end {__file__}')



#%% init project env
import esypro
sfs = esypro.ScriptResultManager('zqf',locals(), version=0)

import torch
from torch import nn
from . import Model, manipulators

class ResConv(Model):
    def __init__(self, in_channels: int, out_channels: int, kernel_size: int = 3, stride: int = 1,
                 stimuli=nn.Hardswish):
        """
        conv block with res connection
        :param in_channels:
        :param out_channels:
        :param kernel_size:
        :param stride:
        """

        super().__init__()
        if isinstance(kernel_size, int):
            padding = kernel_size // 2
        else:
            padding = [k // 2 for k in kernel_size]
        self.res_conv = nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=1, stride=stride ** 2)
        self.conv1 = nn.Sequential(
            nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size, stride=stride,
                      padding=padding),
            stimuli()
        )
        self.conv2 = nn.Sequential(
            nn.Conv2d(in_channels=out_channels, out_channels=out_channels, kernel_size=kernel_size, stride=stride,
                      padding=padding),
            stimuli()
        )

    def get_output(self, x: torch.Tensor) -> torch.Tensor:
        direct = self.conv2(self.conv1(x))
        res = self.res_conv(x)
        return res + direct


class ResTConv(Model):
    def __init__(self, in_channels: int, out_channels: int, kernel_size: int = 3, stride: int = 1,
                 stimuli=nn.Hardswish):
        """
        conv block with res connection
        :param in_channels:
        :param out_channels:
        :param kernel_size:
        :param stride:
        """

        super().__init__()
        self.res_conv = nn.ConvTranspose2d(in_channels=in_channels, out_channels=out_channels, kernel_size=1,
                                           stride=stride ** 2)
        self.conv1 = nn.Sequential(
            nn.ConvTranspose2d(in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size,
                               stride=stride,
                               padding=kernel_size // 2),
            stimuli()
        )
        self.conv2 = nn.Sequential(
            nn.ConvTranspose2d(in_channels=out_channels, out_channels=out_channels, kernel_size=kernel_size,
                               stride=stride,
                               padding=kernel_size // 2),
            stimuli()
        )

    def get_output(self, x: torch.Tensor) -> torch.Tensor:
        direct = self.conv2(self.conv1(x))
        res = self.res_conv(x)
        return res + direct

class MultiplyLayer(manipulators.Multiply):
    def __init__(self, input_dim, output_dim, bias=True):
        route1 = nn.Linear(input_dim, output_dim, bias=bias)
        route2 = nn.Linear(input_dim, output_dim, bias=bias)
        super().__init__(route1, route2)

class FullConnectedLayer(torch.nn.Sequential):
    def __init__(self, input_dim, output_dim, bias=True, stimuli=nn.Hardswish()):
        super().__init__(
            nn.Linear(input_dim, output_dim, bias=bias),
            stimuli
        )


class SparseLinear(torch.nn.Module):
    
    def __init__(self, in_features, out_features, stimuli=torch.nn.ReLU):
        super(SparseLinear, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = torch.nn.Parameter(torch.rand(out_features, in_features))
        self.bias = torch.nn.Parameter(torch.rand(out_features))
        self.stimuli = stimuli()
    
    def forward(self, din: torch.Tensor):
        x = din.to_sparse()
        
        indicies = x.indices()
        values = x.values()
        
        used_weights = self.weight.T
        used_weights = self.weight.unsqueeze(0).expand(x.size(0), -1, -1)
        used_weights = used_weights[indicies[0], indicies[1]]   
        
        used_bias = self.bias
        used_bias = used_bias.unsqueeze(0).expand(x.size(0), -1)
        used_bias = used_bias[indicies[0], indicies[1]]
        
        biased_values = values + used_bias
        
        r_values = (biased_values.unsqueeze(-1).expand(used_weights.size())*used_weights)
        
        batch_sum_r = torch.zeros(x.size(0), self.out_features).to(x.device)
        batch_sum_r = batch_sum_r.index_add(0, indicies[0], r_values)
        
        r = self.stimuli(batch_sum_r)
        return r
        

        
class VAE(Model):
    def __init__(self, encoder, decoder):
        super().__init__()
        self.encoder = encoder
        self.decoder = decoder
        
    def get_output(self, x):
        mu, log_var = self.encode(x)
        z = self.resample(mu, log_var)
        return self.decoder(z)
    
    def encode(self, x):
        mu, log_var = self.encoder(x).chunk(2, dim=1)
        return mu, log_var
    
    def resample(self, mu, log_var):
        z = mu + torch.randn_like(mu) * torch.exp(log_var / 2)
        return z
    
class unsqueeze(torch.nn.Module):
    def __init__(self, dim: int):
        super().__init__()
        self.dim = dim
    
    def get_output(self, x: torch.Tensor):
        return x.unsqueeze(self.dim)
    
class squeeze(torch.nn.Module):
    def __init__(self, dim: int):
        super().__init__()
        self.dim = dim
    
    def get_output(self, x: torch.Tensor):
        return x.squeeze(self.dim)


class SpinConv(Model):
    """
    旋转卷积，目的是使用池化的同时将卷积的维度转移到channel上。 TODO：有点奇怪的是，假如用channel填充被池化的位置，那么，邻域相关性会怎么体现呢
    """
    def __init__(self, in_channels: int, out_channels: int, kernel_size: int = 3, stride: int = 1, padding=None, stimuli=nn.Hardswish, pool_kernel_size: int = 2, pool_padding: int = 0):
        if padding is None:
            padding = kernel_size // 2
        hidden_channels = in_channels * (pool_kernel_size ** 2)
        
        super().__init__()
        
        
        self.conv1 = Model(
            nn.Conv2d(in_channels=in_channels, out_channels=hidden_channels, kernel_size=kernel_size, stride=stride, padding=padding),
            nn.MaxPool2d(kernel_size=pool_kernel_size, stride=pool_kernel_size, padding=pool_padding),
        )
        
        self.conv2 = nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size, stride=stride, padding=padding)
        
        self.conv_res = nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=1, stride=stride ** 2)
        
        self.stimuli = stimuli()
        self.dim_unflatten = Model(
            manipulators.Unflatten(1, (in_channels, pool_kernel_size, pool_kernel_size)),
            manipulators.Permute(((0, 1, 2, 4, 3, 5))),  # b, c, c1, c2, w, h -> b, c, c1, w, c2, h
            manipulators.Flatten(2, 3),  # b, c, c1, w, c2, h -> b, c, c1 * w, c2, h
            manipulators.Flatten(-2, -1)  # b, c, c1 * w, c2, h -> b, c, c1 * w , c2 * h
        )
        
    def get_output(self, x: torch.Tensor) -> torch.Tensor:
        direct = self.conv_res(x)
        
        straight_conv = self.conv1(x)
        
        # flatten channel to the 2d dims
        straight_conv = self.dim_unflatten(straight_conv)  # b, c, c1, c2, w, h -> b, c, c1 * w, c2 * h
        
        straight_conv = self.stimuli(straight_conv)
        
        straight_conv = self.conv2(straight_conv)
        
        straight_conv = self.stimuli(straight_conv)
        
        return direct + straight_conv
    
class SpinConv1d(Model):
    """
    旋转卷积，但是Conv1d，目的是使用池化的同时将卷积的维度转移到channel上。
    """
    def __init__(self, in_channels: int, out_channels: int, kernel_size: int = 3, stride: int = 1, padding=None, stimuli=nn.Hardswish, pool_kernel_size: int = 2, pool_padding: int = 0):
        if padding is None:
            padding = kernel_size // 2
        hidden_channels = in_channels * pool_kernel_size
        
        super().__init__()
        
        self.conv1 = Model(
            nn.Conv1d(in_channels=in_channels, out_channels=hidden_channels, kernel_size=kernel_size, stride=stride, padding=padding),
            nn.MaxPool1d(kernel_size=pool_kernel_size, stride=pool_kernel_size, padding=pool_padding),
        )
        
        self.conv2 = nn.Conv1d(in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size, stride=stride, padding=padding)
        
        self.conv_res = nn.Conv1d(in_channels=in_channels, out_channels=out_channels, kernel_size=1, stride=stride ** 2)
        
        self.stimuli = stimuli()
        self.dim_unflatten = Model(
            manipulators.Unflatten(1, (in_channels, pool_kernel_size)),
            manipulators.Flatten(-2, -1),  # b,c, c1, w -> b,c,c1*w
        )
        
    def get_output(self, x: torch.Tensor) -> torch.Tensor:
        direct = self.conv_res(x)
        
        straight_conv = self.conv1(x)
        
        # flatten channel to the 2d dims
        straight_conv = self.dim_unflatten(straight_conv)  # b,c,c1 -> b,c*c1
        
        straight_conv = self.stimuli(straight_conv)
        
        straight_conv = self.conv2(straight_conv)
        
        straight_conv = self.stimuli(straight_conv)
        
        return direct + straight_conv

#%% main
if __name__ == '__main__':
    print(f'start {__file__}')
    sample_num = 1000

    y = (torch.rand(sample_num) * 5).long()

    x = torch.randn(sample_num, 3, 32)
    
    for i in range(sample_num):
        x_i = x[i]
        y_i = y[i] * 0.01
        x_i[x_i < y_i] = 0
        x[i] = x_i

    #%% analyse data
    # import umap
    # d2_x = umap.UMAP(n_components=2, n_neighbors=10, min_dist=0.1).fit_transform(x.reshape(sample_num, -1))

    # import matplotlib.pyplot as plt
    # plt.scatter(d2_x[:, 0], d2_x[:, 1], c=y, cmap='jet', s=1)
    # plt.colorbar()
    # plt.title('UMAP of x')

    #%% train model for classification
    from .train import TrainningProcess
    from . import basic
    class SimTrainProcess(TrainningProcess):
        def init_dataset(self, record: dict = None):
            total_dataset = basic.Dataset([x, y], name='simulated dataset')
            trainset, valset = total_dataset.k_split(5)[self.val_index]
                
            train_loader = trainset.get_loader(64, num_workers=4, shuffle=False)
            val_loader = valset.get_loader(16, num_workers=4, shuffle=False)
            
            if record is not None:
                record['trainset'] = trainset
                record['valset'] = valset
                record['total_dataset'] = total_dataset
                record['train_loader'] = train_loader
                record['val_loader'] = val_loader
                
            return total_dataset, trainset, valset, train_loader, val_loader
        
        def init_model(self, record: dict = None):
            model = Model(
                SpinConv1d(3, 3, kernel_size=3, stride=2, stimuli=nn.Hardswish, pool_kernel_size=2),
                manipulators.Flatten(1),
                FullConnectedLayer(3 * 8, 5, stimuli=nn.Hardswish()), 
            )
            
            if self.result_dir.cat('model_last.pth').exist():
                model.load_state_dict(torch.load(self.result_dir.cat('model_last.pth'), weights_only=True, map_location=self.device))
            
            if record is not None:
                record['model'] = model
                
            model.to(self.device)
            
            return model

    training_process = SimTrainProcess(sfs.path_of('spin_conv1d').ensure(), device='cuda:0', val_index=0)

    #%%
    training_process.train(3)
        
    
    
    #%% end script
    print(f'end {__file__}')



import esypro
sfs = esypro.ScriptResultManager('zqf', locals())

from . import Model
import torch

class TemporalRes(Model):
    def __init__(self, rnn: torch.nn.RNN,  res_length: int = 1, res_interval: int = None):
        super().__init__()
        self.rnn = rnn
        
        if res_interval is None:
            res_interval = res_length
        
        self.res_length = res_length
        self.res_interval = res_interval
        
    def forward(self, x:torch.Tensor, initial_hidden=None, padding=None):
        """
        :param x: (batch, seq, channels) is a sequential input
        :return: (batch, seq, channels)
        """
        
        # pad sequence
        if padding is None:
            padding = self.res_length - x.size(1) % self.res_length
        x = torch.nn.functional.pad(x, (0, 0, padding, 0))
        
        # seperate x
        x_blocks = x.unfold(1, self.res_length, self.res_interval).permute(1, 0, 3, 2)  # (block, batch, seq, channels)
        
        outputs, hidden = [], []
        
        # first bolck is used to initialize the hidden state
        inital_block = x_blocks[0]
        inital_output, initial_hidden = self.rnn(inital_block, initial_hidden)  # output(batch, seq, channels), hidden(num_layers, batch, channels)
        outputs.append(inital_output)
        res_hidden = initial_hidden * 0
        
        # iterate over the blocks
        for res_block in x_blocks[1:]:
            input_hidden = res_hidden + initial_hidden
            output, hidden = self.rnn(res_block, input_hidden)
            res_hidden = initial_hidden
            initial_hidden = hidden
            outputs.append(output)
        
        # make the output
        outputs = torch.cat(outputs, dim=1)[:, padding:]
        
        
        return outputs, hidden
            
            
            
        

if __name__ == '__main__':

    
    rnn = torch.nn.RNN(
        input_size=10,
        hidden_size=20,
        num_layers=2,
        batch_first=True
    )
    
    seq_length = 20000
    
    sim_inpput = torch.randn(4, seq_length, 10)  # (batch, seq, channels)
    sim_output = torch.randn(2, 4, 20)  # (batch, seq, channels)
    
    model = TemporalRes(rnn, 3)
    simple_rnn = torch.nn.RNN(
        input_size=10,
        hidden_size=20,
        num_layers=2,
        batch_first=True
    )
    loss_fc = torch.nn.MSELoss()
    optimizer = torch.optim.Adam(model.parameters(), lr=0.001)
    optimizer_simple = torch.optim.Adam(simple_rnn.parameters(), lr=0.001)
    
    losses = []
    losses2 = []
    
    for i in range(100):
        output, hidden = model(sim_inpput)
        
        loss = loss_fc(hidden, sim_output)
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        
        output, hidden = simple_rnn(sim_inpput)
        loss2 = loss_fc(hidden, sim_output)
        optimizer_simple.zero_grad()
        loss2.backward()
        optimizer_simple.step()
        
        
        print(f'epoch {i}:',loss.item(), loss2.item())
        losses.append(loss.item())
        losses2.append(loss2.item())
    
    import matplotlib.pyplot as plt
    plt.figure()
    plt.plot(torch.tensor(losses)/max(losses), label='TemporalRes')
    plt.plot(torch.tensor(losses2)/max(losses2), label='SimpleRNN')
    plt.legend()
    plt.savefig(sfs.path_of('loss-norm', 'png').ensure())
    
    



#region utils
from torch.nn.parameter import Parameter
import torch.nn.functional as F
import torch
import torch.nn as nn
import numpy as np


class EqualizedLR_Conv2d(nn.Module):
    def __init__(self, in_ch, out_ch, kernel_size, stride=1, padding=0):
        super().__init__()
        self.padding = padding
        self.stride = stride
        self.scale = np.sqrt(2 / (in_ch * kernel_size[0] * kernel_size[1]))

        self.weight = Parameter(torch.Tensor(out_ch, in_ch, *kernel_size))
        self.bias = Parameter(torch.Tensor(out_ch))

        nn.init.normal_(self.weight)
        nn.init.zeros_(self.bias)

    def forward(self, x):
        return F.conv2d(x, self.weight * self.scale, self.bias, self.stride, self.padding)


class Pixel_norm(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, a):
        b = a / torch.sqrt(torch.sum(a ** 2, dim=1, keepdim=True) + 10e-8)
        return b


class Minibatch_std(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        size = list(x.size())
        size[1] = 1

        std = torch.std(x, dim=0)
        mean = torch.mean(std)
        return torch.cat((x, mean.repeat(size)), dim=1)


#endregion

#region model
from .gans import ProgressiveGAN


#endregion


#region model
class FromRGB(nn.Module):
    def __init__(self, in_ch, out_ch):
        super().__init__()
        self.conv = EqualizedLR_Conv2d(in_ch, out_ch, kernel_size=(1, 1), stride=(1, 1))
        self.relu = nn.LeakyReLU(0.2)

    def forward(self, x):
        x = self.conv(x)
        return self.relu(x)


class ToRGB(nn.Module):
    def __init__(self, in_ch, out_ch):
        super().__init__()
        self.conv = EqualizedLR_Conv2d(in_ch, out_ch, kernel_size=(1, 1), stride=(1, 1))

    def forward(self, x):
        return self.conv(x)


class G_Block(nn.Module):
    def __init__(self, in_ch, out_ch, initial_block=False):
        super().__init__()
        if initial_block:
            self.upsample = None
            self.conv1 = EqualizedLR_Conv2d(in_ch, out_ch, kernel_size=(4, 4), stride=(1, 1), padding=(3, 3))
        else:
            self.upsample = nn.Upsample(scale_factor=2, mode='nearest')
            self.conv1 = EqualizedLR_Conv2d(in_ch, out_ch, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.conv2 = EqualizedLR_Conv2d(out_ch, out_ch, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.relu = nn.LeakyReLU(0.2)
        self.pixelwisenorm = Pixel_norm()
        nn.init.normal_(self.conv1.weight)
        nn.init.normal_(self.conv2.weight)
        nn.init.zeros_(self.conv1.bias)
        nn.init.zeros_(self.conv2.bias)

    def forward(self, x):

        if self.upsample is not None:
            x = self.upsample(x)
        # x = self.conv1(x*scale1)
        x = self.conv1(x)
        x = self.relu(x)
        x = self.pixelwisenorm(x)
        # x = self.conv2(x*scale2)
        x = self.conv2(x)
        x = self.relu(x)
        x = self.pixelwisenorm(x)
        return x


class D_Block(nn.Module):
    def __init__(self, in_ch, out_ch, initial_block=False):
        super().__init__()

        if initial_block:
            self.minibatchstd = Minibatch_std()
            self.conv1 = EqualizedLR_Conv2d(in_ch + 1, out_ch, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
            self.conv2 = EqualizedLR_Conv2d(out_ch, out_ch, kernel_size=(4, 4), stride=(1, 1))
            self.outlayer = nn.Sequential(
                nn.Flatten(),
                nn.Linear(out_ch, 1)
            )
        else:
            self.minibatchstd = None
            self.conv1 = EqualizedLR_Conv2d(in_ch, out_ch, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
            self.conv2 = EqualizedLR_Conv2d(out_ch, out_ch, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
            self.outlayer = nn.AvgPool2d(kernel_size=(2, 2), stride=(2, 2))

        self.relu = nn.LeakyReLU(0.2)
        nn.init.normal_(self.conv1.weight)
        nn.init.normal_(self.conv2.weight)
        nn.init.zeros_(self.conv1.bias)
        nn.init.zeros_(self.conv2.bias)

    def forward(self, x):
        if self.minibatchstd is not None:
            x = self.minibatchstd(x)
            x[torch.isnan(x)] = 0

        x = self.conv1(x)
        x = self.relu(x)
        x = self.conv2(x)
        x = self.relu(x)
        x = self.outlayer(x)
        return x


class ProgressiveGenerator(nn.Module):
    def __init__(self, latent_size, out_res):
        super().__init__()
        self.depth = 1
        self.alpha = 1
        self.fade_iters = 0
        self.upsample = nn.Upsample(scale_factor=2, mode='nearest')
        self.current_net = nn.ModuleList([G_Block(latent_size, latent_size, initial_block=True)])
        self.toRGBs = nn.ModuleList([ToRGB(latent_size, 3)])
        # __add_layers(out_res)
        for d in range(2, int(np.log2(out_res))):
            if d < 6:
                ## low res blocks 8x8, 16x16, 32x32 with 512 channels
                in_ch, out_ch = 512, 512
            else:
                ## from 64x64(5th block), the number of channels halved for each block
                in_ch, out_ch = int(512 / 2 ** (d - 6)), int(512 / 2 ** (d - 5))
            self.current_net.append(G_Block(in_ch, out_ch))
            self.toRGBs.append(ToRGB(out_ch, 3))

    def forward(self, x):
        for block in self.current_net[:self.depth - 1]:
            x = block(x)
        out = self.current_net[self.depth - 1](x)
        x_rgb = self.toRGBs[self.depth - 1](out)
        if self.alpha < 1:
            x_old = self.upsample(x)
            old_rgb = self.toRGBs[self.depth - 2](x_old)
            x_rgb = (1 - self.alpha) * old_rgb + self.alpha * x_rgb

            self.alpha += self.fade_iters

        return x_rgb

    def growing_net(self, num_iters):

        self.fade_iters = 1 / num_iters
        self.alpha = 1 / num_iters

        self.depth += 1


class ProgressiveDiscriminator(nn.Module):
    def __init__(self, latent_size, out_res):
        super().__init__()
        self.depth = 1
        self.alpha = 1
        self.fade_iters = 0

        self.downsample = nn.AvgPool2d(kernel_size=(2, 2), stride=(2, 2))

        self.current_net = nn.ModuleList([D_Block(latent_size, latent_size, initial_block=True)])
        self.fromRGBs = nn.ModuleList([FromRGB(3, latent_size)])
        for d in range(2, int(np.log2(out_res))):
            if d < 6:
                in_ch, out_ch = latent_size, out_res
            else:
                in_ch, out_ch = int(latent_size / 2 ** (d - 5)), int(out_res / 2 ** (d - 6))
            self.current_net.append(D_Block(in_ch, out_ch))
            self.fromRGBs.append(FromRGB(3, in_ch))

    def forward(self, x_rgb):
        x_rgb_ = x_rgb + 0
        x = self.fromRGBs[self.depth - 1](x_rgb)

        x = self.current_net[self.depth - 1](x)
        if self.alpha < 1:
            x_rgb = self.downsample(x_rgb)
            x_old = self.fromRGBs[self.depth - 2](x_rgb)
            x = (1 - self.alpha) * x_old + self.alpha * x
            self.alpha += self.fade_iters
        for block in reversed(self.current_net[:self.depth - 1]):
            x = block(x)

        return torch.sigmoid(x)

    def growing_net(self, num_iters):

        self.fade_iters = 1 / num_iters
        self.alpha = 1 / num_iters

        self.depth += 1

#endregion



#%% init project env
import esypro
sfs = esypro.ScriptResultManager('zqf', locals())

#%% performance for classifying
import torch
def confusion_matrix(target_label:torch.Tensor, output_distribution:torch.Tensor):
    """
    confusion matrix for each layer
    Args:
        target_label (torch.Tensor): target label
        output_distribution (torch.Tensor): output distribution
    Returns:
        confusion_matrix (torch.Tensor): confusion matrix
    """
    # device
    if target_label.device != output_distribution.device:
        target_label = target_label.to(output_distribution.device)
    if len(target_label.size()) > 1:
        target_label = target_label.squeeze()
    assert target_label.size(0) == output_distribution.size(0)
    output_label = output_distribution.argmax(dim=-1)
    num_classes = output_distribution.size(1)
    confusion_matrix = torch.zeros(num_classes, num_classes)
    for i in range(target_label.size(0)):
        confusion_matrix[target_label[i], output_label[i]] += 1
    return confusion_matrix

def accuracy(target_label:torch.Tensor, output_distribution:torch.Tensor):
    """
    accuracy for each layer
    Args:
        target_label (torch.Tensor): target label
        output_distribution (torch.Tensor): output distribution
    Returns:
        accuracy (torch.Tensor): accuracy
    """
    if target_label.device != output_distribution.device:
        target_label = target_label.to(output_distribution.device)
    if target_label.dim() > 1:
        target_label = target_label.squeeze()
    assert target_label.size(0) == output_distribution.size(0)
    output_label = output_distribution.argmax(dim=-1)
    correct_num = (target_label == output_label).sum()
    return correct_num / target_label.size(0)

def precision(confusion_matrix:torch.Tensor):
    """
    precision 
    Args:
        confusion_matrix (torch.Tensor): confusion matrix
    Returns:
        precision (torch.Tensor): precision
    """
    num_classes = confusion_matrix.size(0)
    precisions = torch.zeros(num_classes)
    for i in range(num_classes):
        precisions[i] = confusion_matrix[i, i] / (confusion_matrix[i].sum()+1e-6)
    return precisions



#%% main
if __name__ == '__main__':
    print(f'start {__file__}')
    
    
    
    #%% end script
    



import esypro
sfs = esypro.ScriptResultManager('zqf', locals())

import torch
from . import Model

def hilbert_curve(n):
    def hilbert(x, y, xi, xj, yi, yj, n):
        if n <= 0:
            return [(x + (xi + yi) // 2, y + (xj + yj) // 2)]
        else:
            points = []
            points += hilbert(x, y, yi // 2, yj // 2, xi // 2, xj // 2, n - 1)
            points += hilbert(x + xi // 2, y + xj // 2, xi // 2, xj // 2, yi // 2, yj // 2, n - 1)
            points += hilbert(x + xi // 2 + yi // 2, y + xj // 2 + yj // 2, xi // 2, xj // 2, yi // 2, yj // 2, n - 1)
            points += hilbert(x + xi // 2 + yi, y + xj // 2 + yj, -yi // 2, -yj // 2, -xi // 2, -xj // 2, n - 1)
            return points

    return hilbert(0, 0, 2**n, 0, 0, 2**n, n)


def hilbert_index_matrix(level=4):
    """
    生成希尔伯特整数矩阵
    :param dimensions: 维度
    :param bits: 每维度的位数
    :return: 希尔伯特整数矩阵
    """
    locations = hilbert_curve(level)
    matrix = torch.zeros(2**level, 2**level)
    for i, (x, y) in enumerate(locations):
        matrix[x, y] = i
    return matrix

class HilbertDecoder(Model):
    """
    希尔伯特解码，将输入序列解码为矩阵
    """
    def __init__(self, level=4):
        """
        初始化希尔伯特解码对象，指定级别。作用是将输入序列解码为矩阵
        参数:
            level (int, optional): 希尔伯特曲线的级别。默认值为4。
        属性:
            size (int): 希尔伯特曲线的级别。
            matrix (torch.Tensor): 指定级别的希尔伯特索引矩阵。
            seq_length (int): 序列长度，计算为4的level次方。
        """
        super().__init__()
        self.size = level
        self.matrix = hilbert_index_matrix(level).long()
        self.seq_length = 4**level
    
    def get_output(self, x):
        """
        :param x: [torch.Tensor(seq_l, ...)]输入序列
        """
        # 如果序列长度不够，就自动补全
        if x.size(0) < self.seq_length:
            x = torch.cat(
                [
                    x,
                    torch.zeros(self.seq_length - x.size(0), *x.size()[1:], dtype=x.dtype, device=x.device)
                ]
            )
        
        x = x[self.matrix]
        # 把前两个维度移动到最后，保持其它维度不变
        dim_index = list(range(2, x.dim())) + [0, 1]
        x = x.permute(dim_index)
        # 如果维度个数小于4，则补充channel维度
        if x.dim() < 4:
            x = x.unsqueeze(-3)
        return x


class HilbertEncoder(Model):
    """
    希尔伯特编码，将输入矩阵编码为序列
    """
    def __init__(self, level):
        """
        初始化希尔伯特解码对象，指定级别。作用是将输入矩阵编码为序列
        参数:
            level (int, optional): 希尔伯特曲线的级别。默认值为4。
        属性:
            size (int): 希尔伯特曲线的级别。
            matrix (torch.Tensor): 指定级别的希尔伯特索引矩阵。
            seq_length (int): 序列长度，计算为4的level次方。
        """
        super().__init__()
        self.size = level
        self.matrix = hilbert_index_matrix(level).long()
        self.seq_length = 4**level
        
    def get_output(self, x):
        """
        :param x: [torch.Tensor(..., H, W)]输入矩阵
        """
        # 把后两个维度移动到前面，保持其它维度不变
        dim_index = [-2, -1] + list(range(x.dim() - 2))
        x = x.permute(dim_index)
        
        # 根据矩阵索引，将矩阵转换为序列
        
        y = torch.zeros_like(x).flatten(0, 1)
        
        y[self.matrix] += x
        
        return y
        
# 奇怪，为什么咋样会和1d卷积一样呢？
class HilbertConv(Model):
    def __init__(self, in_channels, out_channels, kernel_level, stride=1, padding=0, dilation=1, groups=1, bias=True):
        """
        希尔伯特卷积，在输入序列上滑窗，将窗口内的序列按照希尔伯特曲线的顺序变成二维矩阵，然后和卷积核相乘，再变回序列
        """
        super().__init__()
        
        self.kernel_size = 2**kernel_level
        self.window_length = 4**kernel_level
        self.kernel_level = kernel_level
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.groups = groups
        self.decoder = HilbertDecoder(kernel_level)
        
        # 生成卷积核
        self.weight = torch.nn.Parameter(torch.Tensor(out_channels, in_channels, self.window_length))
        if bias:
            self.bias = torch.nn.Parameter(torch.Tensor(out_channels))
            

    def forward(self, x):
        """
        :param x: [torch.Tensor(seq_l, batch, in_channels, ...)]输入序列
        """

        # 滑窗，将窗口内的序列按照希尔伯特曲线的顺序变成二维矩阵
        unfolded = x.unfold(0, self.window_length, self.stride)
        unfolded = self.decoder(unfolded.contiguous().T).flatten(-2, -1)  # ..., in_channels, batch, window_length

        # 卷积操作
        output = torch.nn.functional.conv2d(unfolded, self.weight.view(self.out_channels, self.in_channels, self.kernel_size, self.kernel_size), self.bias, self.stride, self.padding, self.dilation, self.groups)

        # 变回序列
        output = output.view(-1, self.out_channels, self.kernel_size * self.kernel_size).permute(0, 2, 1).contiguous().view(-1, self.out_channels)

        return output
        

        


if __name__ == '__main__':
    level = 4

    model = HilbertConv(3, 2, level)
    
    # line = torch.arange(2**(level*2)).view(-1, 1).expand(-1, 3).float()
    line = torch.rand(1000, 4, 3, 2)
    
    r = model(line)
    
    
    



import esypro
sfs = esypro.ScriptResultManager('zqf', locals())

import torch
import esypro

def max_level_of(dataset_path: esypro.MyPath, max_search=10):
    """
    获取dataset_path下样本的最大深度
    """
    max_level = 1
    current_path = dataset_path
    while max_level < max_search:
        files = current_path.get_files(list_r=True)
        if len(files) > 0:
            break
        current_path = current_path / max(list(current_path.get_files('').keys()))
        max_level += 1
    return max_level

def sample_index_to_path(max_level:int, sample_index:int, code=100):

    # 获取每一级的索引
    path = []
    for i in range(max_level-1):
        path.append(sample_index % code)
        sample_index = sample_index // code
    path.append(sample_index)
    path.reverse()  
    
    # 转换为路径
    p = '/'.join([str(p) for p in path])
    return p
        
def path_to_sample_index(path, code=100):
    """
    将路径转换为样本索引
    """
    path = path.split('/')

    sample_index = 0
    for i, p in enumerate(path):
        sample_index += int(p) * (code ** (len(path) - i - 1))
    return sample_index

def save_samples_in_dataset(dataset_path, samples, indexs:torch.Tensor=None, suffix='.pt', max_level=None, code=100, bar=False):
    """
    将样本保存到文件系统中，每个样本是一个文件，文件名是样本的索引
    由于文件系统子文件夹数量有限，所以子文件夹下最多有100个对象，每100个对象则构建更深一层目录（其实就是按照100进制存储）
    """
    if indexs is None:
        indexs = torch.arange(len(samples))
        
    if max_level is None:
        max_level = 1
        while indexs.max().item() > code ** max_level:
            max_level += 1

    if bar:
        from tqdm import tqdm
        bar = tqdm(zip(indexs, samples), total=len(indexs))
    else:
        bar = zip(indexs, samples)
    for idx, sample in bar:
        p =  dataset_path / (sample_index_to_path(max_level, idx.item(), code) + suffix)
        # 确保文件夹存在，由于可能需要适配多线程，所以使用try
        try:
            p.ensure()
        except:
            pass
        
        torch.save(sample, p)

def multi_process_save(dst_path: esypro.MyPath, iterable_obj: object, indexs: list=None, suffix:str='.pt', max_level: int=None, code: int=100, bar: bool=False):
    """
    多进程保存
    Args:
        dst_path (esypro.MyPath): 保存路径
        iterable_obj (object): 可迭代对象
        indexs (list): 索引列表, 每个子进程处理一个索引
        suffix (str, optional): 后缀. Defaults to '.pt'.
        max_level (int, optional): 最大深度. Defaults to None.
        code (int, optional): 每层最大数量. Defaults to 100.
        bar (bool, optional): 是否显示进度条. Defaults to False.
    """
    total_indexs = torch.arange(len(iterable_obj))
    
    # 把total_index转换为dask array
    from dask import array as dask_array
    dasked_index = dask_array.from_array(total_indexs.numpy(), chunks=(code,))
    
    dasked_index.map_blocks(lambda x: save_samples_in_dataset(dst_path, [iterable_obj[i] for i in x], x, suffix, max_level, code, bar), dtype='object').compute()
    

def max_index_in(folder: esypro.MyPath):
    """
    获取文件夹下最大的索引
    """
    files = folder.get_files(list_r=False)
    if len(files) == 0:
        return -1
    return max([int(f.split('.')[0]) for f in files.keys()])

def last_sample_path(dataset_path, max_level=None, code=100):
    """
    获取dataset_path下最后一个样本的路径
    """
    if max_level is None:
        max_level = max_level_of(dataset_path)
    
    current_path = dataset_path
    for level in range(max_level):
        files = current_path.get_files(list_r=True)
        if len(files) == 0:  # 如果当前目录下没有文件，则进入最大的子目录
            current_path = current_path / str(max([int(n) for n in list(current_path.get_files('', list_r=False).keys())]))
        else:  # 如果当前目录下有文件，则输出最大的文件路径
            path = current_path / str(max_index_in(current_path))
            
    return path.relative_to(dataset_path)[1:]

from . import Dataset
class FileDataset(Dataset):
    def __init__(self, dataset_path, sample_selection=None, suffixs=['.pt'], code=100):
        # 获取最大深度
        max_level = max_level_of(dataset_path)
        
        if sample_selection is None:
            # 获取最大样本的路径
            last_sample = last_sample_path(dataset_path, max_level)
            # 获取最大样本的索引
            max_index = path_to_sample_index(last_sample)
            sample_selection = torch.arange(max_index+1)
        tensors = [sample_selection]
        super().__init__(tensors, sample_selection)
        self.dataset_path = dataset_path
        self.max_level = max_level
        self.code = code
        self.suffixs = suffixs
        
    def __getitem__(self, select_index):
        index = super().__getitem__(select_index)[0]
        path = self.dataset_path / sample_index_to_path(self.max_level, index.item(), self.code)
        result = []
        for suffix in self.suffixs:
            result.append(torch.load(path + suffix))
        return result

if __name__ == '__main__':
    import torch
    data_index = torch.arange(150)
    
    dataset_path = sfs.path_of('dataset', '').ensure()
    
    save_samples_in_dataset(dataset_path, data_index, suffix='.pt')
    
    print(max_level_of(dataset_path))
    
    print(last_sample_path(dataset_path))
    
    dataset = FileDataset(dataset_path)
    
    for batch_result in dataset.get_loader(batch_size=4):
        print(batch_result)
        break
    
    
        


import torch
import visdom
from tqdm import tqdm

from . import Model


class RandomLatent(Model):
    def __init__(self, latent_dim=100):
        super().__init__()
        self.latent_dim = latent_dim
        self.d_p = torch.nn.Parameter(torch.randn(latent_dim))

    def forward(self, batch_size=4):
        return torch.randn(batch_size, self.latent_dim).to(self.device)


class Generator(Model):
    def __init__(self, latent_dim=100, *layers):
        rand_layer = RandomLatent(latent_dim)
        super().__init__(*[rand_layer, *layers])
        self.latent_dim = rand_layer.latent_dim


def get_din(batch):
    return batch[-1]


class GAN(Model):
    def __init__(self, generator, discriminator):
        super().__init__()
        self.generator = generator
        self.discriminator = discriminator
        self.generator_optimizer = None
        self.discriminator_optimizer = None
        self.loss_function = torch.nn.BCELoss()

    def fit(
            self, data_loader, n_cretic=1, messages: dict = None, get_din=get_din, bar: tqdm = None,
            sample_internal=0, viz: visdom.Visdom = None
    ):
        if messages is None:
            messages = {}
            for loss_name in 'train_batch_losses_g, train_batch_losses_d, test_batch_losses_g, test_batch_losses_d'.replace(
                    ' ', '').split(','):
                messages[loss_name] = []
                if viz is not None:
                    viz.line([0], [0], win=loss_name, opts=dict(title=loss_name))

        if sample_internal > 0 and not ('example_sample' in messages.keys()):
            messages['example_sample'] = []

        #region training prepare
        if self.generator_optimizer is None:
            self.generator_optimizer = torch.optim.Adam(self.generator.parameters(), lr=0.0002, betas=(0.5, 0.999))
        if self.discriminator_optimizer is None:
            self.discriminator_optimizer = torch.optim.Adam(self.discriminator.parameters(), lr=0.0002,
                                                            betas=(0.5, 0.999))
        self.loss_function = self.loss_function.to(self.device)
        #endregion

        for batch in data_loader:
            step = bar.n
            bar.update(1)

            #region din
            din = get_din(batch).to(self.device)
            batch_size = din.size(0)
            #endregion

            #region labels
            real_label = torch.zeros(batch_size, 1).to(self.device)
            fake_label = torch.ones(batch_size, 1).to(self.device)
            #endregion

            #region train discriminator
            real_judge = self.discriminator(din)
            real_loss = self.loss_function(real_judge, real_label)

            fake = self.generator(batch_size)
            fake_judge = self.discriminator(fake)
            fake_loss = self.loss_function(fake_judge, fake_label)

            train_batch_loss_d = real_loss + fake_loss

            self.discriminator_optimizer.zero_grad()
            train_batch_loss_d.backward()
            self.discriminator_optimizer.step()

            messages['train_batch_losses_d'].append(train_batch_loss_d.item())
            try:
                if viz is not None:
                    viz.line([train_batch_loss_d.item()], [step], win='train_batch_losses_d', update='append')
            except:
                pass
            #endregion

            #region train generator
            if step % n_cretic == 0:
                generate_result = self.generator(batch_size)
                generate_judge = self.discriminator(generate_result)
                generate_loss = self.loss_function(generate_judge, real_label)

                self.generator_optimizer.zero_grad()
                generate_loss.backward()
                self.generator_optimizer.step()

                messages['train_batch_losses_g'].append(generate_loss.item())
                try:
                    if viz is not None:
                        viz.line([generate_loss.item()], [step], win='train_batch_losses_g', update='append')
                except:
                    pass
                if step % sample_internal == 0:
                    messages['example_sample'].append(generate_result[0].detach().cpu())
                    try:
                        if viz is not None:
                            viz.image(messages['example_sample'][-1], win=f'example_sample{step}')
                    except:
                        pass
            #endregion

        return messages


class PGGLoss(torch.nn.Module):
    def forward(self, p1, p2):
        samples, fake, D_net, lambd = p2
        eps = torch.rand(samples.size(0), 1, 1, 1)
        eps = eps.expand_as(samples)
        x_hat = eps * samples + (1 - eps) * fake.detach()
        x_hat.requires_grad = True
        px_hat = D_net(x_hat)
        grad = torch.autograd.grad(
            outputs=px_hat.sum(),
            inputs=x_hat,
            create_graph=True
        )[0]
        grad_norm = grad.view(samples.size(0), -1).norm(2, dim=1)
        gradient_penalty = lambd * ((grad_norm - 1) ** 2).mean()

        return gradient_penalty

class ProgressiveGAN(GAN):
    """
    need generator and discriminator with depth differed output
    """
    depth_iter_num = 0

    def __init__(self, generator, discriminator):
        super().__init__(generator, discriminator)

        self.loss_function = torch.nn.L1Loss()

    def fit(
            self, data_loader, n_cretic=1, messages: dict = None, get_din=get_din, bar: tqdm = None,
            sample_internal=0, viz: visdom.Visdom = None, grow=False
    ):
        def resize(img):
            img = get_din(img)
            assert len(img.size()) == 4
            return torch.nn.functional.interpolate(img, size=2 ** (self.discriminator.depth + 1))

        if grow:
            self.grow()
        r = super().fit(data_loader, n_cretic, messages, resize, bar, sample_internal, viz)

        self.depth_iter_num += len(data_loader.dataset)

        return r

    def grow(self):
        self.generator.layer2.growing_net(self.depth_iter_num)
        self.discriminator.growing_net(self.depth_iter_num)



import kan.KANLayer
import torch
from . import Model


class KAN_Conv(Model):
    def __init__(self, in_channels=3, out_channels=1, kernel_size=3, stride=1, padding=1, dilation=1, groups=1, bias=True):
        super(KAN_Conv, self).__init__()
        self.conv = torch.nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size, stride=stride, padding=padding, dilation=dilation, groups=groups, bias=bias)
        self.kan1 = kan.KANLayer(out_channels, out_channels)

    def forward(self, input):
        output = self.conv(input).transpose(-3, -1)
        size_rec = output.shape[:-1]
        output = self.kan1(output.flatten(0, -2))[0].unflatten(0, size_rec).transpose(-3, -1)
        return output
        

class KAN_GRU(Model):
    def __init__(self, input_size=2, hidden_size=5):
        super(KAN_GRU, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.kan1 = kan.KANLayer(input_size+hidden_size, hidden_size)
        self.kan2 = kan.KANLayer(input_size+hidden_size, hidden_size)
        self.kan3 = kan.KANLayer(input_size+hidden_size, hidden_size)

    def forward(self, input, initial_state):
        prev_h = initial_state
        bs, T, i_size = input.shape

        output = torch.zeros(bs, T, h_size)

        for t in range(T):
            r_t = self.kan1(torch.cat((input[:, t, :], prev_h), dim=-1))[0]
            z_t = self.kan2(torch.cat((input[:, t, :], prev_h), dim=-1))[0]
            n_t = self.kan3(torch.cat((input[:, t, :], r_t*prev_h), dim=-1))[0]
            h_t = (1 - z_t) * n_t + z_t * prev_h
            prev_h = h_t
            output[:, t, :] = prev_h # 将最新状态的hidden_state移动到输出矩阵中

        return output, prev_h

if __name__ == '__main__':

    # #region test KAN GRU

    # #定义常量
    # bs, T, i_size, h_size = 2, 3, 4, 5
    # input = torch.randn(bs, T, i_size) #输入序列
    # h0 = torch.randn(bs, h_size) #初始值,不需要训练
    # target = torch.rand(bs, T, h_size) #目标序列

    # #定义模型
    # model = KAN_GRU(i_size, h_size)

    # # train prepare
    # loss_f = torch.nn.L1Loss()

    # opt = torch.optim.Adam(model.parameters(), lr=0.01)

    # # train

    # for i in range(500):
    #     output, h = model(input, h0)
    #     opt.zero_grad()
    #     loss = torch.nn.functional.mse_loss(output, target)
    #     loss.backward()
    #     opt.step()
    #     print(loss.item())
    #     # optimizer.step()
    # #endregion

    #region test KAN Conv
    b, c, w, h = 5, 3, 28, 28
    input = torch.randn(b, c, w, h)
    target = torch.zeros(b)
    target[:b//2] = 1
    target = target.long()

    model = torch.nn.Sequential(
        KAN_Conv(3, 1, 3),
        torch.nn.Flatten(1),
        torch.nn.Linear(28*28, 2)
    )

    loss_f = torch.nn.CrossEntropyLoss()

    opt = torch.optim.Adam(model.parameters(), lr=0.01)

    for i in range(500):
        output = model(input)
        opt.zero_grad()
        loss = loss_f(output, target)
        loss.backward()
        opt.step()
        print(loss.item())
        # optimizer.step()
    #endregion


#%%
import EsyPro
sfs = EsyPro.ScriptFileSaver(__file__, locals())

import torch
from EsyMod import Dataset, SerialDataset
from scipy.sparse import csr_matrix
import h5py

class ParalellDtaset(Dataset):
    def __init__(self, datasets:Dataset, sample_selection=None):
        super().__init__(datasets, sample_selection)

class H5DenseSample(Dataset):
    def __init__(self, h5_path:str, X_path:str='raw/obs/cell_type/codes'):
        """
        一个H5数据集
        """
        self.h5_path = h5_path
        self.X_path = X_path
        super().__init__([torch.arange(len(self.Xs)),])
        print(self)
        
    def __getitem__(self, index):
        index = super().__getitem__(index)[0]
        x_values = self.Xs[index]
        
        return x_values
    
    @ property
    def h5(self):
        return h5py.File(self.h5_path, 'r')
    
    @ property
    def Xs(self):
        try:
            r= self.h5[self.X_path]
        except:
            print(f'err when get {self.X_path} from {self.h5_path}')
        return r

import numpy as np
class H5StrSample(H5DenseSample):
    def __init__(self, h5_path: str, X_path: str = 'raw/obs/cell_type/codes'):
        super().__init__(h5_path, X_path)
        
        categories = np.array(self.Xs)
        categories = np.unique(categories)
        self.categories = categories.tolist()
        
    def __getitem__(self, index):
        r = super().__getitem__(index)
        return self.categories.index(r)
    



class H5SparseSample(ParalellDtaset):
    def __init__(self, h5_path:str, X_path:str='raw/X/data', indices_path:str='raw/X/indices', indptr_path:str='raw/X/indptr', data_shape=(1, 70000)):
        """
        一个H5数据集
        """
        self.h5_path = h5_path
        self.X_path = X_path
        self.indices_path = indices_path
        self.indptr_path = indptr_path
        self.data_shape = data_shape
        
        super().__init__([torch.arange(len(self.indptrs)-1),])
    
    @property
    def h5(self):
        return h5py.File(self.h5_path, 'r')
    
    @property
    def Xs(self):
        return self.h5[self.X_path]
    @property
    def indices(self):
        return self.h5[self.indices_path]
    @property
    def indptrs(self):
        return self.h5[self.indptr_path]
        
    def __getitem__(self, index):
        index = super().__getitem__(index)[0]
        x_values = self.Xs[self.indptrs[index]:self.indptrs[index+1]]
        x_indexs = self.indices[self.indptrs[index]:self.indptrs[index+1]]
        x_pointers = self.indptrs[index:index+2] - self.indptrs[index]
        
        data_sparse = csr_matrix((x_values, x_indexs, x_pointers), self.data_shape)
        
        return data_sparse
    
    def to_dense(self):
        dataset = DensedH5SparseSample(self.h5_path, self.X_path, self.indices_path, self.indptr_path, self.data_shape)
        dataset.sample_selection = self.sample_selection
        return dataset
    
class DensedH5SparseSample(H5SparseSample):
    def __getitem__(self, index)->torch.Tensor:
        r = super().__getitem__(index)
        return torch.tensor(r.toarray())



def mv_sparse_data(src_h5_path, dst_h5_path, selection):
    assert len(selection.shape) == 1  # now only support 1D selection
    
    with h5py.File(src_h5_path, 'r') as src_file:
        X_data = src_file['raw/X/data']
        X_indices = src_file['raw/X/indices']
        X_indptr = src_file['raw/X/indptr']
    
        write_selected_sparse(X_data, X_indices, X_indptr, selection, dst_h5_path)
    
    return True

from EsyImg.process_bar import rich_tqdm as tqdm
def write_selected_sparse(X_data, X_indices, X_indptr, selection, dst_h5_path):

    data_num = 0
    for i, select in enumerate(selection):
        if select:
            data_num += X_indptr[i+1] - X_indptr[i]
    
    with h5py.File(dst_h5_path, 'w') as dst_file:
        dst_file.require_dataset('raw/X/data', shape=(data_num,), dtype='float32')
        dst_file.require_dataset('raw/X/indices', shape=(data_num,), dtype='int64')
        dst_file.require_dataset('raw/X/indptr', shape=(selection.sum()+1,), dtype='int64')
        
        dst_file['raw/X/indptr'][0] = 0
        select_indptr_end = 0
        last_indptr = 0
        n = 0
        for i, select in tqdm(enumerate(selection), total=len(selection)):
            if select:
                select_data = X_data[X_indptr[i]:X_indptr[i+1]]
                select_indices = X_indices[X_indptr[i]:X_indptr[i+1]]
                select_indptr_end = last_indptr+select_data.shape[0]
                dst_file['raw/X/data'][last_indptr:select_indptr_end] = select_data
                dst_file['raw/X/indices'][last_indptr:select_indptr_end] = select_indices
                n += 1         
                dst_file['raw/X/indptr'][n] = select_indptr_end
                
                last_indptr = select_indptr_end


# %%
if __name__ == '__main__':
    print(f'start {__file__}')

    #%%
    print(f'end {__file__}')


import torch
import torch.utils
import torch.utils.data

class CheckOut(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.output = None
    def forward(self, x):
        self.output = x
        return x

class Model(torch.nn.Sequential):
    """
    1. 带有设备参数的模型，用于设备自动适配。
    2. 带有保存和加载方法的模型，用于模型持久化。
    3. 带有参数数量属性的模型，用于模型参数计数。
    4. 带有输出属性的模型，用于前向输出。
    """
    
    def __init__(self, *layers):
        """
        使用层初始化模型。
        """
        super().__init__()
    
        self.device_parameter = torch.nn.Parameter(torch.tensor(0, dtype=torch.int64), requires_grad=False)
        for idx, module in enumerate(layers):
            self.add_module(f'layer{idx}', module)
    
    def get_output(self, x):
        return super().forward(x)
        
    def forward(self, *args, **kwargs):
        """
        前向传播

        参数:
            x (Tensor): 输入张量

        返回:
            Tensor: 输出张量
        """
        self.output = self.get_output(*args, **kwargs)
        return self.output

    #region device
    @ property
    def device(self):
        return self.device_parameter.device
        
    
    #endregion

    #region param count
    @ property
    def param_num(self):
        params = list(self.parameters())
        k = 0
        for i in params:
            l = 1
            for j in i.size():
                l *= j
            k = k + l
        return k
    #endregion

import copy
class IterableData(torch.utils.data.Dataset):
    """
    一个自定义数据集类，继承自 `torch.utils.data.Dataset`，用于处理张量数据并提供额外的功能，如迭代、数据加载和k折交叉验证。 \n
    **属性**:\n
        sample_selection (torch.Tensor): 包含要使用的样本索引的张量。\n
        tensors (list of torch.Tensor): 包含数据的张量列表。\n
    **方法**:\n
        __init__(self, tensors, sample_selection=None):\n
            使用给定的张量和可选的样本选择初始化数据集。\n
        __iter__(self):\n
            初始化数据集的迭代器。\n
        __next__(self):\n
            在迭代过程中返回数据集中的下一个项目。\n
        get_loader(self, batch_size=4, shuffle=True):\n
            返回具有指定批量大小和随机选项的数据加载器。\n
        __getitem__(self, select_index):\n
            返回指定索引处的项目。\n
        __len__(self):\n
            返回数据集的长度。\n
        subset(self, sample_selection):\n
            创建一个具有新样本选择的数据集副本。\n
        k_split(self, k):\n
            将数据集分成k折进行交叉验证，并返回(train_subset, val_subset)元组的列表。
    """
    def __init__(self, iterable_obj, sample_selection=None, name=None):
        """
        使用给定的张量和可选的样本选择初始化数据集。
        参数:
            tensors (list of torch.Tensor): 包含数据的张量列表。
            sample_selection (torch.Tensor): 包含要使用的样本索引的张量。
        """
        if sample_selection is None:
            sample_selection = torch.arange(len(iterable_obj))
        if name is None:
            name = type(self).__name__
        
        self.sample_selection = sample_selection
        self.iterable_obj = iterable_obj
        self.name = name
        

    #region iterable
    def __iter__(self):
        self.iter_count = -1
        return self

    def __next__(self):
        self.iter_count += 1
        if self.iter_count >= len(self):
            raise StopIteration
        return self[self.iter_count]
    #endregion
    
    def __repr__(self) -> str:
        return self.name + f': {len(self)} samples'

    def get_loader(self, batch_size=4, shuffle=True, **args):
        """
        返回具有指定批量大小和随机选项的数据加载器。
        参数:
            batch_size (int): 数据加载器的批量大小。
            shuffle (bool): 是否随机打乱数据。
        返回:
            (DataLoader): 数据加载器。
        """
        return torch.utils.data.DataLoader(self, batch_size=batch_size, shuffle=shuffle, **args)

    def __getitem__(self, select_index):
        
        index = self.sample_selection[select_index].item()
        return self.iterable_obj[index]

    def __len__(self):
        return len(self.sample_selection)

    def subset(self, sample_selection: torch.Tensor|torch.BoolTensor):
        """
        拷贝一个共享所有属性的数据集，单独替换sample_selection
        参数:
            sample_selection (torch.Tensor): 包含要使用的样本索引的张量。
        返回:
            (Dataset): 具有新样本选择的数据集副本。
        """
        # 如果sample_selection是truefalse向量，转换为索引
        if sample_selection.dtype == torch.bool:
            sample_selection = self.sample_selection[sample_selection]
        
        copyset = copy.copy(self)
        copyset.sample_selection = sample_selection
        return copyset
    
    def k_split(self, k:int, shuffle:bool=True):
        """
        按照k折划分验证的要求，返回交叉验证数据集.
        参数:
            k (int): 交叉验证的折数。
        返回:
            (Dataset, Dataset): (train_subset, val_subset)元组的列表。
        """
        total_index = self.sample_selection
        # shuffle
        if shuffle:
            total_index = total_index[torch.randperm(len(total_index))]
        
        # split
        subsets = []
        for i in range(k):
            val_index = total_index[i::k]
    
            train_index = total_index[~torch.isin(total_index, val_index)]        
            val_subset = self.subset(val_index)
            train_subset = self.subset(train_index)
            subsets.append((train_subset, val_subset))
        
        return subsets
        
    
class Dataset(IterableData):
    """
    一个自定义数据集类，继承自 `torch.utils.data.Dataset`，用于处理张量数据并提供额外的功能，如迭代、数据加载和k折交叉验证。 \n
    **属性**:\n
        sample_selection (torch.Tensor): 包含要使用的样本索引的张量。\n
        tensors (list of torch.Tensor): 包含数据的张量列表。\n
    **方法**:\n
        __init__(self, tensors, sample_selection=None):\n
            使用给定的张量和可选的样本选择初始化数据集。\n
        __iter__(self):\n
            初始化数据集的迭代器。\n
        __next__(self):\n
            在迭代过程中返回数据集中的下一个项目。\n
        get_loader(self, batch_size=4, shuffle=True):\n
            返回具有指定批量大小和随机选项的数据加载器。\n
        __getitem__(self, select_index):\n
            返回指定索引处的项目。\n
        __len__(self):\n
            返回数据集的长度。\n
        subset(self, sample_selection):\n
            创建一个具有新样本选择的数据集副本。\n
        k_split(self, k):\n
            将数据集分成k折进行交叉验证，并返回(train_subset, val_subset)元组的列表。
    """
    def __init__(self, tensors, sample_selection=None, **kwargs):
        """
        使用给定的张量和可选的样本选择初始化数据集。
        参数:
            tensors (list of torch.Tensor): 包含数据的张量列表。
            sample_selection (torch.Tensor): 包含要使用的样本索引的张量。
        """
        
        if sample_selection is None:
            sample_selection = torch.arange(len(tensors[0]))
        
        self.tensors = []
        for tensor in tensors:
            if not isinstance(tensor, IterableData):
                tensor = IterableData(tensor, sample_selection)
            self.tensors.append(tensor)
        
        super().__init__(sample_selection, sample_selection, **kwargs) 
            

    def __getitem__(self, select_index):
        """返回指定索引处的项目。
        参数:
            select_index (int): 要获取的项目的索引。
        返回:
            list: 包含所有张量在指定索引处的值的列表。[index, tensor1_value, tensor2_value, ...]
            
        """
        
        index = self.sample_selection[select_index].item()
        result = [index]
        for tensor in self.tensors:
            result.append(tensor[index])
        return result


class SerialDataset(Dataset):
    def __init__(self, datasets: list[Dataset], sample_selection=None):
        """
        数据集中必须有一个是SerialDataset，可以通过以下方式构建:
        [
            Dataset(
                [
                    Dataset(xxx),
                    Dataset(xxx)
                ]
            )
        ]
        """
        self.datasets = datasets
        dataset_indexs = []
        inner_indexs = []
        for i, dataset in enumerate(datasets):
            d_index = torch.tensor([i]*len(dataset))
            i_index = torch.arange(len(dataset))
            dataset_indexs.append(d_index)
            inner_indexs.append(i_index)
        dataset_indexs = torch.cat(dataset_indexs)
        inner_indexs = torch.cat(inner_indexs)        
        super().__init__([dataset_indexs, inner_indexs])
        
    def __getitem__(self, select_index):
        index, data_index, inner_index = super().__getitem__(select_index)
        return index, self.datasets[data_index][inner_index]

    def get_hiarchical_index(self, select_index):
        """
        获取高阶索引
        参数:
            select_index (int): 选择的索引
        返回:
            (int, int): (dataset_index, inner_index)
        """
        index, data_index, inner_index = super().__getitem__(select_index)
        return data_index.item(), inner_index.item()


# -*- coding:utf-8 -*-
import sys
sys.argv.append('sdist')
from distutils.core import setup
from setuptools import find_packages

setup(name='esymod',
            version='39.2025.7.16.17.58',
            packages=['esymod',],
            description='a python lib for project files',
            long_description='',
            author='Quanfa',
            include_package_data = True,
            author_email='quanfa@tju.edu.cn',
            url='http://www.xxxxx.com/',
            license='MIT',
            )

            


#%% init project env
import esypro
sfs = esypro.ScriptResultManager('zqf',locals(), version=0)
import torch

# 如果有esyimg就导入esyimg，否则导入tqdm
rich_tqdm_available = False
try:
    from esyimg.process_bar import rich_tqdm as tqdm
    from esyimg.process_bar import progress
    rich_tqdm_available = True
except ImportError:
    from tqdm import tqdm

class TrainningProcess:
    rich_tqdm_available = rich_tqdm_available
    def __init__(self, result_dir:esypro.MyPath|str, val_index=0, device='cuda:0'):
        if isinstance(result_dir, str):
            self.result_dir = esypro.MyPath(result_dir)
        
        self.val_index = val_index
        self.device = device
        self.result_dir = result_dir.ensure()
    
    def continue_train_of(self, result_dir:esypro.MyPath|str):
        # 复制result_dir下的所有文件到self.result_dir下
        import shutil
        shutil.copytree(result_dir, self.result_dir, dirs_exist_ok=True)
    
    def init_dataset(self, record:dict=None):
        
        if not self.result_dir.cat('dataset.pth').exist():
            
            from .basic import Dataset
            total_dataset = Dataset([torch.randn(100, 10), torch.randint(0, 10, (100,))])
            trainset, valset = total_dataset.k_split(5)[self.val_index]

            torch.save((total_dataset, trainset, valset), self.result_dir.cat('dataset.pth'))
        else:
            total_dataset, trainset, valset = torch.load(self.result_dir.cat('dataset.pth'))
        
        train_loader = trainset.get_loader(64, num_workers=4, shuffle=False)
        val_loader = valset.get_loader(16, num_workers=4, shuffle=False)
        
        if record is not None:
            record['trainset'] = trainset
            record['valset'] = valset
            record['total_dataset'] = total_dataset
            record['train_loader'] = train_loader
            record['val_loader'] = val_loader
            
        return total_dataset, trainset, valset, train_loader, val_loader
    
    def init_model(self, record:dict=None):
        
        from .blocks import FullConnectedLayer
        model = FullConnectedLayer(10, 10).to(self.device)
        
        if self.result_dir.cat('model_last.pth').exist():
            model.load_state_dict(torch.load(self.result_dir.cat('model_last.pth'), weights_only=True, map_location=self.device))
        
        torch.save(model.state_dict(), self.result_dir.cat('model_last.pth'))
        
        if record is not None:
            record['model'] = model
        
        return model
    
    def init_optimizer(self, record):
        from torch.optim import SGD
        model = record['model']
        optimizer = SGD(model.parameters(), lr=1e-3)
        record['optimizer'] = optimizer
        return optimizer
    
    def batch_step(self, record):
        batch_data = record['batch_data']
        model = record['model']
        index, x, y = batch_data
        x, y = x.to(self.device), y.to(self.device)
        y_pred = model(x)
        loss = torch.nn.functional.cross_entropy(y_pred, y)
        
        batch_result = {'model_output':y_pred, 'loss':loss}
        
        record['batch_result'] = batch_result
        
        return batch_result
    
    def record_init(self):
        record = {}
        
        self.init_dataset(record)
        self.init_model(record)        
        self.init_optimizer(record)
        
        return record
    
    def after_batch_description(self, record):
        return f'epoch{record["epoch"]}, train lossL{record["train_batch_result"]["loss"]}, val loss:{record["val_batch_result"]["loss"]}'
    
    def after_batch(self, record):
        description = self.after_batch_description(record)
        if self.rich_tqdm_available:
            progress.update(progress.task_ids[-1], description=description)
        else:
            print(description)
    
    def after_epoch(self, record):
        print('finish epoch',record['epoch'],'======================================')
        
        # save model
        torch.save(record['model'].state_dict(), self.result_dir.cat(f'model_last.pth'))
        
    
    def train(self, epoch_num=5):
        record = self.record_init()

        # train
        for epoch in tqdm(range(epoch_num)):
            record['epoch'] = epoch
            for train_batch, val_batch in tqdm(zip(record['train_loader'], record['val_loader']), total=len(record['train_loader'])):

                # train
                record['batch_data'] = train_batch
                train_batch_result = self.batch_step(record)
                record['train_batch_result'] = train_batch_result
                record['optimizer'].zero_grad()
                train_batch_result['loss'].backward()
                record['optimizer'].step()
                                
                # val
                with torch.no_grad():
                    record['batch_data'] = val_batch
                    val_batch_result = self.batch_step(record)
                    record['val_batch_result'] = val_batch_result
                    
                self.after_batch(record)
                
            self.after_epoch(record)
                    
        return record

class BatchStepProcess:
    @ staticmethod
    def simple_forward(model, batch_data, device='cuda:0'):
        """
        A simple forward step for a model, used in training or validation.
        Args:
            model (torch.nn.Module): The model to be evaluated.
            batch_data (tuple): A tuple containing the input data and labels.
            device (str): The device to run the model on, default is 'cuda:0'.
        Returns(dict):
            model_output (torch.Tensor): The output of the model.
            loss (torch.Tensor): The computed loss.
        """
        x, y = batch_data
        x, y = x.to(device), y.to(device)
        y_pred = model(x)
        loss = torch.nn.functional.cross_entropy(y_pred, y)
        return {'model_output': y_pred, 'loss': loss}
    
    @ staticmethod
    def denoise_step(model, original_data, noise_step=1, device='cuda:0'):
        """
        A denoise step for a model, used in training or validation.
        Args:
            model (torch.nn.Module): The model to be evaluated.
            original_data (torch.Tensor): The original data to be denoised.
            step (int): The step number.
            device (str): The device to run the model on, default is 'cuda:0'.
        Returns(dict):
            denoised_output (torch.Tensor): The denoised output of the model.
        """
        alpha_t = 1-torch.exp(-noise_step - 1)
        alpha_t_ =  1-torch.exp(-noise_step)
        
        original_data = original_data.to(device)
        alpha_t_ = alpha_t_.to(device)
        alpha_t = alpha_t.to(device)
        
        size = [alpha_t.size(0), *original_data.size()]
        
        noise_data_current = (torch.sqrt(alpha_t).view(-1, 1, 1, 1).expand(size))*(original_data.unsqueeze(0).expand(size)) + (torch.sqrt(1-alpha_t).view(-1, 1, 1, 1).expand(size))* (torch.randn_like(original_data).unsqueeze(0).expand(size))
        denoised_data_next = (torch.sqrt(alpha_t_).view(-1, 1, 1, 1).expand(size))*(original_data.unsqueeze(0).expand(size)) + (torch.sqrt(1-alpha_t_).view(-1, 1, 1, 1).expand(size))* (torch.randn_like(original_data).unsqueeze(0).expand(size))
        
        target_output = denoised_data_next - noise_data_current
        
        # batch flatten and channel add
        noise_data_current = noise_data_current.flatten(0, 1).unsqueeze(1)
        target_output = target_output.flatten(0, 1).unsqueeze(1)
        
        denoised_output = model(noise_data_current)
        
        if False:
            from esyimg import web_matplotlib_util as plt
            figure = plt.figure(figsize=(15, 10))
            plt.imshow(denoised_output[0, 0].cpu().detach().numpy(), cmap='gray')
            plt.show()
        
        loss = torch.nn.functional.mse_loss(denoised_output, target_output)
        return {'denoised_output': denoised_output, 'loss': loss}
    
    
                
                
        

#%% main
if __name__ == '__main__':
    print(f'start {__file__}')

    
    
    #%% end script
    print(f'end {__file__}')



import torch.utils
import torch.utils.data
import esypro
sfs = esypro.ScriptResultManager('zqf', locals())

"""

"""
from esymod import Model
import torch
from .blocks import FullConnectedLayer
    
        

class BaseLayer(Model):
    def __init__(self, taylor_order: int=3, work_points: int=3, input_dim: int=10, output_dim: int=10):
        """
        
        """
        super(BaseLayer, self).__init__()
        self.w_model = torch.nn.Linear(input_dim*taylor_order*work_points, output_dim*taylor_order*work_points)
        self.alpha_model = torch.nn.Linear(input_dim*taylor_order*work_points, output_dim*taylor_order*work_points)
        self.taylor_order = taylor_order
        self.work_points = work_points
        
        
        
    def forward(self, x):
        """
        forward the model.
        """
        # construct din
        din = []
        for i in range(self.taylor_order):
            din.append(x**i)
        din = torch.cat(din, dim=-1)
        din = din.unsqueeze(-2).expand(-1, 3, -1).flatten(-2)
        
        # forward
        wx = self.w_model(din)
        alpha = self.alpha_model(din)
        return wx*alpha

        
        


if __name__ == '__main__':
    
    import torch
    
    dataset = torch.utils.data.TensorDataset(torch.randn(100, 100000).cuda(2), torch.randn(100, 10).cuda(2))
    
    data_loader = torch.utils.data.DataLoader(dataset, batch_size=100)
    
    model = torch.nn.Sequential(
        torch.nn.Linear(100000, 10000),
        torch.nn.ReLU(),
        torch.nn.Linear(10000, 1000),
        torch.nn.ReLU(),
        torch.nn.Linear(1000, 10),
    ).cuda(2)
    
    for i in range(10000):
        for x, y in data_loader:
            y_pred = model(x)
            loss = torch.nn.functional.mse_loss(y_pred, y)
            loss.backward()
            print(loss)
    
    
    
    



import esypro
sfs = esypro.ScriptResultManager('zqf', locals())
import torch
from torch.nn import Identity
from . import Model


class Squeeze(torch.nn.Module):
    def __init__(self, dim=None):
        super().__init__()
        self.dim = dim

    def forward(self, x):
        return torch.squeeze(x, dim=self.dim)
    
class Unsqueeze(torch.nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.dim = dim

    def forward(self, x):
        return torch.unsqueeze(x, dim=self.dim)

class Flatten(torch.nn.Flatten):
    def __init__(self, start_dim=0, end_dim=-1):
        super().__init__(start_dim, end_dim)
        self.original_shape = None
        self.output_shape = None
    def forward(self, x):
        self.original_shape = tuple(x.shape)
        r = super().forward(x)
        self.output_shape = tuple(r.shape)
        return r

class Unflatten(torch.nn.Module):
    def __init__(self, dim, sizes):
        super().__init__()
        self.dim = dim
        self.sizes = sizes
        
    def forward(self, x):
        sizes = self.sizes
        if isinstance(sizes, Flatten):
            sizes = sizes.original_shape[sizes.start_dim:sizes.end_dim+1]
        
        return torch.unflatten(x, self.dim, sizes)

class AlignStack(torch.nn.Module):
    def __init__(self, align_dim=1, target_size=0):
        """
        align tensors to the same size on align_dim, and stack them on 0 dim
        :param stack_dim: the dim to stack
        :param align_dim: the dim to align, index orignally
        :param target_size: the target size of align_dim
        """
        super().__init__()
        self.align_dim = align_dim
        self.target_size = target_size
        
    def forward(self, tensors: list[torch.Tensor]):
        target_size = self.target_size
        if target_size == 0:
            target_size = max([tensor.shape[self.align_dim] for tensor in tensors])
        
        example_size = tensors[0].shape
        example_size = list(example_size)
        example_size[self.align_dim] = target_size
        example_size = [len(tensors)] + example_size
        result = torch.zeros(example_size, device=tensors[0].device)
        
        for i, tensor in enumerate(tensors):
            slices = [slice(None)] * len(tensor.shape)
            slices[self.align_dim] = slice(None, tensor.shape[self.align_dim])
            slices = [slice(i)] + slices
            result[slices] = tensor

        return result

class View(torch.nn.Module):
    def __init__(self, *shape):
        super().__init__()
        self.shape = shape
        self.original_shape = []
        
    def forward(self, x):
        self.original_shape[0] = x.shape.copy()
        id = 0
        shape = []
        for s in self.shape:
            if s == -1:
                shape.append(x.size(id))
                id += 1
            else:
                shape.append(s)
        return x.view(*shape)

class PadUnflatten(torch.nn.Module):
    def __init__(self, dim, size):
        """
        unflatten a dim to size, pad if not match
        """
        super().__init__()
        self.dim = dim
        self.size = size
        self.ideal_size = torch.tensor(size).cumprod(0)[-1].item()
    
    def forward(self, x):
        if not x.size(self.dim) == self.ideal_size:
            pad_size = self.ideal_size - x.size(self.dim)
            if self.dim == -1:
                pad = torch.zeros(*x.size()[:self.dim], pad_size, device=x.device)
            else:
                pad = torch.zeros(*x.size()[:self.dim], pad_size, *x.size()[self.dim+1:], device=x.device)
            x = torch.cat([x, pad], dim=self.dim)
        return torch.unflatten(x, self.dim, self.size)

class Permute(torch.nn.Module):
    def __init__(self, *dims):
        super().__init__()
        self.dims = dims
    
    def forward(self, x):
        return x.permute(*self.dims)

class Transpose(torch.nn.Module):
    def __init__(self, dim0, dim1):
        super().__init__()
        self.dim0 = dim0
        self.dim1 = dim1
    
    def forward(self, x):
        return x.transpose(self.dim0, self.dim1)

class OutputSelect(torch.nn.Module):
    def __init__(self, selector=0):
        """
        here is an example:
        import torch

        # 假设你有一个tensor
        tensor = torch.randn(3, 5, 7)

        # 使用slice对象
        selector = (slice(None), 4, slice(None))
        
        """
        super().__init__()
        self.selector = selector
        
    def forward(self, x):
        self.output = x[self.selector]
        return self.output

class Stack(torch.nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.dim = dim
    
    def forward(self, xs):
        return torch.stack(xs, dim=self.dim)

class Concatenate(torch.nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.dim = dim
    
    def forward(self, xs):
        return torch.cat(xs, dim=self.dim)

#region math
class SUM(Model):
    def __init__(self, *routes):
        super().__init__(*routes)
        self.routes = routes

    def manipulate(self, x1, x2):
        return x1 + x2

    def forward(self, x):
        rs = [route(x) for route in self.routes]
        result = rs[0]
        for r in rs[1:]:
            result = self.manipulate(result, r)
        self.output = result
        return result


class Multiply(SUM):
    def manipulate(self, x1, x2):
        return x1 * x2


class MatrixMultiply(Multiply):
    def manipulate(self, x1, x2):
        return x1 @ x2


class Exp(torch.nn.Module):
    def forward(self, x):
        return torch.exp(x)
#endregion



# hilbert
class HilbertSelect(torch.nn.Module):
    def __init__(self, level):
        super().__init__()
        self.level = level
        self.hilbert_matrix = self.create_hilbert_matrix(level)
        
    def forward(self, x):
        """
        :param x: a tensor with shape (..., length)
        """
        # trans sequecne to matrix using hilbert matrix
        return x[..., self.hilbert_matrix]
    
    @ staticmethod
    def create_hilbert_matrix(level):
        """
        创建一个矩阵，这个矩阵中的数值为希尔伯特曲线上的index
        """
        return
    

if __name__ == '__main__':

    matrix_depth = 3
    line = torch.arange(2**matrix_depth*2**matrix_depth*1).reshape(1, -1)
    
    model = HilbertSelect(matrix_depth)
    print(model(line))
    
    

